﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace MontoyaTech.Parallel.Net
{
    /// <summary>
    /// The outline of a ParallelTask that takes in an input, and produces and output when ran.
    /// </summary>
    /// <typeparam name="InputT"></typeparam>
    /// <typeparam name="ResultT"></typeparam>
    public class ParallelTask<InputT, ResultT>
    {
        /// <summary>
        /// Whether or not this parallel task is currently running.
        /// </summary>
        public bool Running
        {
            get
            {
                return this.running && !this.completed && !this.failed;
            }
        }

        /// <summary>
        /// Whether or not this parallel task completed successfully.
        /// </summary>
        public bool Completed
        {
            get
            {
                return this.completed;
            }
        }

        /// <summary>
        /// Whether or not this parallel task failed due to an exception.
        /// </summary>
        public bool Failed
        {
            get
            {
                return this.failed;
            }
        }

        /// <summary>
        /// The result of this parallel task if it completed successfully.
        /// </summary>
        public ResultT Result
        {
            get
            {
                return this.result;
            }
        }

        /// <summary>
        /// The exception that was thrown while running this parallel task if it failed.
        /// </summary>
        public Exception ThrownException
        {
            get
            {
                return this.thrownException;
            }
        }

        protected bool running = false;

        protected bool completed = false;

        protected bool failed = false;

        protected Func<InputT, ResultT> work = null;

        protected InputT input = default(InputT);

        protected ResultT result = default(ResultT);

        protected bool exceptionStopped = false;

        protected ManualResetEvent handle;

        protected Exception thrownException = null;

        /// <summary>
        /// Creates a new parallel task with the worker function and input.
        /// </summary>
        /// <param name="work"></param>
        /// <param name="input"></param>
        public ParallelTask(Func<InputT, ResultT> work, InputT input)
        {
            this.work = work;

            this.input = input;
        }

        /// <summary>
        /// Runs this parallel task and returns the result when it completes.
        /// </summary>
        /// <param name="timeout">Default is null, if set, this function will return if the timeout is reached regardless if the task is complete.</param>
        /// <param name="blocking">Default is true, whether or not to block the caller function and not return until this task is complete.</param>
        /// <param name="throwOnTimeout">Whether or not to throw an exception if a timeout is reached, if one was specified. Default is true.</param>
        /// <param name="countDown">An optional count down that if specified is signaled when the task completes.</param>
        /// <returns>The result from the worker function ran in this task.</returns>
        /// <exception cref="Exception">Throws a timeout exception if the task doesn't complete in time.</exception>
        public ResultT Run(TimeSpan? timeout = null, bool blocking = true, bool throwOnTimeout = true, CountdownEvent countDown = null)
        {
            this.running = true;

            this.handle = new ManualResetEvent(false);

            ThreadPool.QueueUserWorkItem((object state) =>
            {
                try
                {
                    this.result = this.work((InputT)state);

                    this.completed = true;
                }
                catch (Exception ex)
                {
                    this.failed = true;

                    this.thrownException = ex;
                }

                this.handle.Set();

                if (countDown != null)
                    countDown.Signal();

            }, this.input);

            if (blocking)
            {
                if (!timeout.HasValue)
                    this.handle.WaitOne();
                else if (!this.handle.WaitOne(timeout.Value) && throwOnTimeout)
                    throw new Exception($"ParallelTask did not complete within {timeout.Value.TotalMilliseconds} miliseconds");
            }

            return this.result;
        }
    }
}
